from Bio.Phylo.PhyloXML import Clade, Phylogeny

def clades_equal(
    clade_a: Clade, clade_b: Clade,
    check_branch_lengths: bool = True
) -> bool:
    if check_branch_lengths and clade_a.branch_length != clade_b.branch_length:
        return False
    if len(clade_a.clades) != len(clade_b.clades):
        return False
    if len(clade_a.clades) == 0:
        return clade_a.name == clade_b.name
    return all([
        clades_equal(
            clade_a.clades[subclade_index],
            clade_b.clades[subclade_index],
            check_branch_lengths=check_branch_lengths
        )
        for subclade_index in range(len(clade_a))
    ])      

def phylogenies_equal(
    phylogeny_a: Phylogeny, phylogeny_b: Phylogeny,
    check_branch_lengths: bool = True
) -> bool:
    return clades_equal(
        phylogeny_a.root, phylogeny_b.root,
        check_branch_lengths=check_branch_lengths
    )


from dataclasses import dataclass
from typing import Optional
from Bio.Phylo.BaseTree import Tree, Clade
from Bio.Phylo.PhyloXML import Phylogeny
from Bio.SeqFeature import SeqFeature
from .featureUtils import SeqFeaturesByContiguity
from .hor import HORInSeq, hor_tree_as_phyloxml_phylogeny, loop_to_HOR, name_hor_tree
from .hor_coherence import checkLoopInSeqSelfOverlap, checkSpanListSelfOverlap
from .loops import find_loops, loop_to_spans
from .phylogeny_to_levels import extract_features_from_leaves, phylogeny_to_levels

def levels_to_hor_tree(
    labelled_items_by_level: list[list[int]],
    clades_by_level: list[list[Clade]],
    seq_features: list[SeqFeature],
    gap_indices: list[int],
    min_loop_size: int = 1,
    max_loop_size: int = 30,
    min_loops: int = 3,
    allowed_mismatch_rate: float = 0.0,
    allow_hor_overlap: bool = False
) -> HORInSeq:
    seq_locations = [seq_feature.location for seq_feature in seq_features]
        
    def level_to_hors(
        spans_to_search: list[tuple[int,int]],
        level: int,
        curr_loop_diversity: int,
        min_loop_size: int
    ):
        # print(f"Level: {level}")
        if level == 0:
            return []
        # checkSpanListSelfOverlap(
        #     spanList=spans_to_search,
        #     foundOverlapTemplate=(
        #         f"In level {level}, " +
        #         "found overlap in searching spans between {span1} and {span2}"
        #     )
        # )
        # print(f"Labelled items: {labelled_items_by_level[level]}")
        # print(f"Spans to search: {spans_to_search}")
        loops_in_seq = find_loops(
            whole_seq=labelled_items_by_level[level],
            seq_spans=spans_to_search,
            min_loop_size=min_loop_size if level < len(labelled_items_by_level) - 1 else 1,
            max_loop_size=max_loop_size,
            min_loops=min_loops,
            allowed_mismatch_rate=allowed_mismatch_rate,
            allow_overlap=allow_hor_overlap
        )
        # print(f"Loops found: {[str(loop_in_seq) for loop_in_seq in loops_in_seq]}")
        # for loop_in_seq_index, loop_in_seq in enumerate(loops_in_seq):
        #     checkLoopInSeqSelfOverlap(
        #         loopInSeq=loop_in_seq,
        #         foundOverlapTemplate=(
        #             f"In level {level}, loop n. {loop_in_seq_index}, " +
        #             "found overlap between {loopSpanInSeq1} and {loopSpanInSeq2}"
        #         )
        #     )
        if (len(loops_in_seq) == 0):
            return []
        
        if (
            len(loops_in_seq) == 1 and
            len(set(loops_in_seq[0].loop.loop_seq)) <= curr_loop_diversity
        ):
            return level_to_hors(
                spans_to_search=loop_to_spans(loops_in_seq[0]),
                level=level - 1,
                curr_loop_diversity=curr_loop_diversity,
                min_loop_size=min_loop_size
            )
        
        hors_in_seq = []
        for loop_in_seq in loops_in_seq:
            hor_in_seq = loop_to_HOR(
                loop_in_seq,
                clades=clades_by_level[level],
                seq_locations=seq_locations
            )
            hor_in_seq.sub_hors = level_to_hors(
                spans_to_search=loop_to_spans(loop_in_seq),
                level=level - 1,
                curr_loop_diversity=len(set(loop_in_seq.loop.loop_seq)),
                min_loop_size=len(loop_in_seq.loop.loop_seq)                
            )
            hors_in_seq.append(hor_in_seq)
            
        return hors_in_seq
    
    split_limits = [0] + gap_indices + [len(seq_locations)]
    num_splits = len(gap_indices) + 1
    hor_tree_root = level_to_hors(
        spans_to_search=[
            (split_limits[split_index], split_limits[split_index + 1])
            for split_index in range(num_splits)
        ],
        level=len(labelled_items_by_level) - 1,
        curr_loop_diversity=0,
        min_loop_size=min_loop_size
    )[0]
    name_hor_tree(hor_tree_root)
    return hor_tree_root

@dataclass
class PhylogenyToHorTreeResult:
    as_hor_in_seq: HORInSeq
    as_phyloxml: Phylogeny

def phylogeny_to_hor_tree(
    phylogeny: Tree,
    max_allowed_gap: int = 10,
    min_loop_size: int = 1,
    max_loop_size: int = 50,
    min_loops: int = 5,
    allowed_mismatch_rate: float = 0.0,
    allow_hor_overlap: bool = False,
    discrete_sorted_levels: bool = False,
    set_branch_lengths: Optional[bool] = None
) -> PhylogenyToHorTreeResult:
    """Given a set of related DNA/RNA sequences, occurring in contiguous
    blocks and called monomers, this function looks for higher order
    repeats, i.e. repeats of sequence of families of sequences. 
    It expects an arleady built phylogenetic tree of the monomers and a 
    set of optional parameters.

    The function has been designed with the purpose of analysing the
    structure of repetitive patterns in centromeres, but there is no
    reason it cannot be used in other contexts.

    Parameters
    ----------
    phylogeny : Bio.Phylo.BaseTree.Tree
        Phylogenetic tree of the monomers. Leaves must have positional
        information of the corresponding sequence, either as attached
        Sequence object in the case of PhyloXML trees (suggested) or
        encoded in the leaf name using the <ref_seq>:<start>-<end>
        format (this second option does not currently support
        positioning in the negative strand).
    max_allowed_gap: int, default=10
        Maximum accepted gap (in number of bases) between monomers
        considered contiguous for HOR discovery purposes.
    min_loop_size: int, default=1
        Minimum size of the loop of recognized HORs (unless you know
        what your are doing, keep this with value one to get the full
        tree, including monomeric regions).
    max_loop_size: int, default=30,
        Maximum loop length of recognized HORs.
    min_loops: int = 5,
        Minimum number of repetitions required for identifying as an HOR.
    allowed_mismatch_rate: float, default=0.0
        Max allowed mismatch rate when identitying an HOR an its
        coverage.
    allow_hor_overlap: bool, default=False
        If True, multiple overlapping can be considered.
        If False, the overlapping parts are assigned to one of the HORs.
    discrete_sorted_levels: bool, default=False
        If True, HORs are searched monomer-tree-level-wise where levels
        are defined by height from leaves, without considering branch
        lengths.
        If False, HORs are searched monomer-tree-level-wise considering
        branch lengths.
    set_branch_lengths: bool, default=not discrete_sorted_levels
        If True, branch lenghts for the HOR tree are set in the
        generated PhyloXML, based on branch lengths in the monomer tree.
        If False, branch lengths are not set.
        Setting to True is meaningful only if discrete_sorted_levels is
        set to False.

    Returns
    -------
    PhylogenyToHorTreeResult:
        - as_hor_in_seq: HORInSeq
            The HORInSeq instance that is the root of the HOR tree.
        - as_phyloxml: Phylogeny
            A PhyloXML phylogeny representating the HOR tree.
    """
    
    sfbc = SeqFeaturesByContiguity(
            seq_features=extract_features_from_leaves(phylogeny),
            max_allowed_gap=max_allowed_gap
    )
    levels_res = phylogeny_to_levels(
        phylogeny=phylogeny,
        item_position_to_leaf_index=sfbc.reordered_indices,
        discrete_sorting=discrete_sorted_levels
    )
    hor_tree = levels_to_hor_tree(
        labelled_items_by_level=levels_res.labelled_items_by_level,
        clades_by_level=levels_res.clades_by_level,
        seq_features=sfbc.sorted_seq_features,
        gap_indices=sfbc.gap_indices,
        min_loop_size=min_loop_size,
        max_loop_size=max_loop_size,
        min_loops=min_loops,
        allowed_mismatch_rate=allowed_mismatch_rate,
        allow_hor_overlap=allow_hor_overlap
    )
    
    if set_branch_lengths is None:
        set_branch_lengths = not discrete_sorted_levels
    
    return PhylogenyToHorTreeResult(
        as_hor_in_seq=hor_tree,
        as_phyloxml=hor_tree_as_phyloxml_phylogeny(
            hor_tree,
            set_branch_lengths=set_branch_lengths
        )
    )
    
def find_inversion_hors(
    seq_features: list[SeqFeature],
    max_allowed_gap: int = 10,
    min_loop_size: int = 2,
    max_loop_size: int = 30,
    min_loops: int = 5,
    allowed_mismatch_rate: float = 0.0,
    allow_hor_overlap: bool = False
) -> HORInSeq:
    sfbc = SeqFeaturesByContiguity(
            seq_features=seq_features,
            max_allowed_gap=max_allowed_gap,
            independent_strands=False
    )
    split_limits = [0] + sfbc.gap_indices + [len(seq_features)]
    num_splits = len(sfbc.gap_indices) + 1
    return levels_to_hor_tree(
        labelled_items_by_level=[
            sfbc.reordered_indices,
            [(1 - seq_feature.strand) // 2 for seq_feature in sfbc.sorted_seq_features],
            [0 for seq_feature in sfbc.sorted_seq_features]
        ],
        clades_by_level=[
            seq_features,
            [Clade(name="+"),Clade(name="-")],
            [Clade(name=".")]
        ],
        seq_features=sfbc.sorted_seq_features,
        gap_indices=sfbc.gap_indices,
        min_loop_size=min_loop_size,
        max_loop_size=max_loop_size,
        min_loops=min_loops,
        allowed_mismatch_rate=allowed_mismatch_rate,
        allow_hor_overlap=allow_hor_overlap
    )
    return find_loops(
        whole_seq=[(1 - seq_feature.strand) // 2 for seq_feature in sfbc.sorted_seq_features],
        seq_spans=[
            (split_limits[split_index], split_limits[split_index + 1])
            for split_index in range(num_splits)
        ],
        min_loop_size=min_loop_size,
        max_loop_size=max_loop_size,
        min_loops=min_loops,
        allowed_mismatch_rate=allowed_mismatch_rate,
        allow_overlap=allow_hor_overlap
    )   


from collections.abc import Iterable
import numpy as np
from Bio.Phylo.PhyloXML import Clade, Phylogeny, Sequence, Phyloxml
from Bio.SeqFeature import SeqFeature


def new_clade(
    label: str = None,
    branch_length: float = None,
    clades: list[Clade] = None
) -> Clade:
    if clades is not None:
        if branch_length is not None:
            for clade in clades:
                clade.branch_length += branch_length
        if len(clades) == 1:
            return clades[0]
    return Clade(name=label, clades=clades, branch_length=0)


def merge_clades(
    clades: Iterable[Clade],
    new_clusters_matrix: np.ndarray,
    branch_length:float = None
) -> list[Clade]:
    return [
        new_clade(
            clades=[
                clade
                for index, clade in enumerate(clades)
                if new_cluster_row[index] > 0
            ],
            branch_length=branch_length
        )
        for new_cluster_row in new_clusters_matrix
    ]


def new_leaves(leaf_labels: Iterable[str]) -> list[Clade]:
    return [new_clade(label=leaf_label) for leaf_label in leaf_labels]


def feature_to_leave(feature: SeqFeature, branch_length: int = 0) -> Clade:
    return Clade(
        name=feature.id,
        branch_length=branch_length,
        sequences=[Sequence(type='dna', location=feature.location)])


def features_to_leaves(features: Iterable[SeqFeature]) -> list[Clade]:
    return [feature_to_leave(feature) for feature in features]


def new_phylogeny(root_clade: Clade, name: str = 'monomers') -> Phylogeny:
    return Phylogeny(root=root_clade, name=name)


def new_phyloXML(phylogenies: list[Phylogeny], attributes: dict={}) -> Phyloxml:
    return Phyloxml(attributes, phylogenies=phylogenies)



import math
import numpy as np

def matrix_to_triu(matrix: np.ndarray) -> np.ndarray:
    return matrix[np.triu_indices_from(matrix, k=1)]

def triu_to_matrix(triu: np.ndarray) -> np.ndarray:
    matrix_size = int((math.sqrt(1 + 8 *len(triu)) + 1) / 2)
    matrix = np.zeros((matrix_size, matrix_size), dtype=triu.dtype)
    matrix[np.triu_indices_from(matrix, k=1)] = triu
    return matrix + matrix.T

def save_matrix_as_triu(matrix: np.ndarray, filename: str):
    with open(filename, 'wb') as f:
        np.save(f, matrix_to_triu(matrix))
    
def load_matrix_from_triu(filename: str) -> np.ndarray:
    return triu_to_matrix(np.load(filename))



from Bio.SeqFeature import SeqFeature
from .featureUtils import SeqFeaturesByContiguity
from .loops import LoopInSeq, find_loops

def find_inversion_loops(
    seq_features: list[SeqFeature],
    max_allowed_gap: int = 10,
    min_loop_size: int = 2,
    max_loop_size: int = 30,
    min_loops: int = 5,
    allowed_mismatch_rate: float = 0.0,
    allow_hor_overlap: bool = False
) -> list[LoopInSeq]:
    sfbc = SeqFeaturesByContiguity(
            seq_features=seq_features,
            max_allowed_gap=max_allowed_gap,
            independent_strands=False
    )
    split_limits = [0] + sfbc.gap_indices + [len(seq_features)]
    num_splits = len(sfbc.gap_indices) + 1
    return find_loops(
        whole_seq=[(1 - seq_feature.strand) // 2 for seq_feature in sfbc.sorted_seq_features],
        seq_spans=[
            (split_limits[split_index], split_limits[split_index + 1])
            for split_index in range(num_splits)
        ],
        min_loop_size=min_loop_size,
        max_loop_size=max_loop_size,
        min_loops=min_loops,
        allowed_mismatch_rate=allowed_mismatch_rate,
        allow_overlap=allow_hor_overlap,
        min_diversity=2
    )    



from functools import partial
import multiprocessing
import math
import numpy as np
import editdistance
from Bio.SeqRecord import SeqRecord
from multiprocessing import Pool
from dataclasses import dataclass

num_bytes_to_max_value_map = [
    2**(num_bytes * 8) - 1 for num_bytes in range(1, 33)
]


def num_bytes_to_max_value(num_bytes: int) -> int:
    return num_bytes_to_max_value_map[num_bytes - 1]

def limited(x: int, num_bytes: int) -> int:
    return min(x, num_bytes_to_max_value(num_bytes))

def num_bytes_to_uint_dtype(num_bytes: int) -> np.dtype:
    if num_bytes == 1:
        return np.uint8
    if num_bytes == 2:
        return np.uint16
    if num_bytes <= 4:
        return np.uint32
    if num_bytes <= 8:
        return np.uint64
    if num_bytes <= 16:
        return np.uint128
    if num_bytes <= 32:
        return np.uint256
    raise Exception('Unsupported num of bytes requested for integer')

def build_string_distance_triu(
    strings: list[str],
    num_bytes_for_each_distance: int = 1
) -> np.ndarray:
    seq_triu = np.array(
        [
            (
                0 if j <= i
                else limited(
                    editdistance.eval(seq_i, strings[j]),
                    num_bytes=num_bytes_for_each_distance
                )
            )
            for i, seq_i in enumerate(strings)
            for j in range(len(strings))
        ],
        dtype=num_bytes_to_uint_dtype(num_bytes_for_each_distance)
    )
    seq_triu.shape = (len(strings), len(strings))
    return seq_triu


def build_string_distance_matrix(
    strings: list[str],
    num_bytes_for_each_distance: int = 1
) -> np.ndarray:
    seq_triu = build_string_distance_triu(strings, num_bytes_for_each_distance)
    return seq_triu + seq_triu.T


def build_string_cross_distance_matrix(
    row_strings: list[str],
    col_strings: list[str],
    num_bytes_for_each_distance: int = 1
) -> np.ndarray:
    dist_matrix = np.array(
        [
            limited(
                editdistance.eval(row_string, col_string),
                num_bytes=num_bytes_for_each_distance
            )
            for row_string in row_strings
            for col_string in col_strings
        ],
        dtype=num_bytes_to_uint_dtype(num_bytes_for_each_distance)
    )
    dist_matrix.shape = (len(row_strings), len(col_strings))
    return dist_matrix


def build_seqs_distance_matrix(
    seqs: list[SeqRecord],
    num_bytes_for_each_distance: int = 1
) -> np.ndarray:
    return build_string_distance_matrix(
        [str(seq.seq) for seq in seqs],
        num_bytes_for_each_distance=num_bytes_for_each_distance
    )

@dataclass
class ChunkIndex:
    row: int
    col: int

class ChunkParams:
    def get_index(self) -> ChunkIndex:
        pass

@dataclass
class ChunkParamsInternal(ChunkParams):
    row_index: int
    col_index: int
    row_strings: list[str]
    col_strings: list[str]

    def __str__(self):
        return f"[{self.row_index}:{self.col_index}]({len(self.row_strings)},{len(self.col_strings)})"
    
    def get_index(self):
        return ChunkIndex(self.row_index, self.col_index)


@dataclass
class ChunkParamsDiagonal(ChunkParams):
    index: int
    strings: list[str]

    def __str__(self):
        return f"[{self.index}:{self.index}]({len(self.strings)},{len(self.strings)})"

    def get_index(self):
        return ChunkIndex(self.index, self.index)

class ChunkResults:

    def get_data(self) -> np.ndarray:
        pass


@dataclass
class ChunkResultsInternal(ChunkResults):
    row_index: int
    col_index: int
    dist_matrix: np.ndarray

    def get_data(self) -> np.ndarray:
        return self.dist_matrix

@dataclass
class ChunkResultsDiagonal(ChunkResults):
    index: int
    dist_triu: np.ndarray

    def get_data(self) -> np.ndarray:
        return self.dist_triu



@dataclass
class JobParams:
    chunk_params: list[ChunkParams]

    def __str__(self):
        return "(" + ",".join([str(cp) for cp in self.chunk_params]) + ")"


@dataclass
class JobResult:
    chunk_results: list[ChunkResults]


class ChunkStore:
    def get(self, chunk_params: ChunkParams) -> ChunkResults:
        return None

    def set(self, chunk_params: ChunkParams, chunk_results: ChunkResults):
        pass

dummy_chunk_store = ChunkStore()

class FileSystemChunkStore:
    filename_template: str
    
    def __init__(self, filename_template) -> None:
        self.filename_template = filename_template
    
    def get(self, chunk_params: ChunkParams) -> ChunkResults:
        filename = self.filename_template.format_map(vars(chunk_params.get_index()))
        try:
            cache_chunk_results_data = np.load(filename + '.npy')
            if isinstance(chunk_params, ChunkParamsDiagonal):
                return ChunkResultsDiagonal(
                    index=chunk_params.index,
                    dist_triu=cache_chunk_results_data
                )
            if isinstance(chunk_params, ChunkParamsInternal):
                return ChunkResultsInternal(
                    row_index=chunk_params.row_index,
                    col_index=chunk_params.col_index,
                    dist_matrix=cache_chunk_results_data
                )
        except IOError:
            return None

    def set(self, chunk_params: ChunkParams, chunk_results: ChunkResults):
        filename = self.filename_template.format_map(vars(chunk_params.get_index()))
        np.save(filename, chunk_results.get_data())


def compute_chunk(
    chunk_params: ChunkParams,
    num_bytes_for_each_distance: int = 1
) -> ChunkResults:
    if isinstance(chunk_params, ChunkParamsDiagonal):
        return ChunkResultsDiagonal(
            index=chunk_params.index,
            dist_triu=build_string_distance_triu(
                chunk_params.strings,
                num_bytes_for_each_distance)
        )
    if isinstance(chunk_params, ChunkParamsInternal):
        return ChunkResultsInternal(
            row_index=chunk_params.row_index,
            col_index=chunk_params.col_index,
            dist_matrix=build_string_cross_distance_matrix(
                row_strings=chunk_params.row_strings,
                col_strings=chunk_params.col_strings,
                num_bytes_for_each_distance=num_bytes_for_each_distance
            )
        )

def compute_chunk_if_needed(
    chunk_params: ChunkParams,
    chunk_store: ChunkStore = dummy_chunk_store,
    num_bytes_for_each_distance: int = 1
) -> ChunkResults:
    cached_chunk_results = chunk_store.get(chunk_params)
    if cached_chunk_results is not None:
        return cached_chunk_results
    fresh_chunk_results = compute_chunk(
        chunk_params,
        num_bytes_for_each_distance
    )
    chunk_store.set(chunk_params, fresh_chunk_results)
    return fresh_chunk_results

def execute_job(
    job_params: JobParams,
    chunk_store: ChunkStore = dummy_chunk_store,
    num_bytes_for_each_distance: int = 1
) -> JobResult:
    return JobResult([
        compute_chunk_if_needed(
            chunk_params, chunk_store,
            num_bytes_for_each_distance
        )
        for chunk_params in job_params.chunk_params
    ])


def build_string_distance_matrix_by_chunks(
    strings: list[str],
    num_chunks: int = None,
    max_num_processes: int = None,
    chunk_store: ChunkStore = dummy_chunk_store,
    num_bytes_for_each_distance = 1
) -> np.ndarray:

    num_strings = len(strings)

    if (num_chunks is None):
        if (max_num_processes is None):
            print(f"CPUs detected: {multiprocessing.cpu_count()}")
            max_num_processes = multiprocessing.cpu_count()
        num_chunks = int(math.sqrt(max_num_processes * 2))

    print(f"# of chunks for computing distance matrix: {num_chunks}")

    chunk_size = math.ceil(num_strings / num_chunks)

    print(f"Chunk size: {chunk_size}")
    
    num_chunks = math.ceil(num_strings / chunk_size)

    def chunk(index):
        offset = index * chunk_size
        return strings[offset:offset + chunk_size]

    internal_blocks = [
        ChunkParamsInternal(
            row_strings=chunk(row_index),
            col_strings=chunk(col_index),
            row_index=row_index,
            col_index=col_index
        )
        for col_index in range(num_chunks)
        for row_index in range(col_index)
    ]

    diagonal_blocks = [
        ChunkParamsDiagonal(
            strings=chunk(index),
            index=index
        )
        for index in range(num_chunks)
    ]

    jobs_params = [
        JobParams([internal_block])
        for internal_block in internal_blocks
    ] + [
        JobParams(diagonal_blocks[i*2:(i+1)*2])
        for i in range(math.ceil(len(diagonal_blocks) / 2))
    ]

    print(f"Num processes: {len(jobs_params)}")
    print(f"Blocks: {[str(jb) for jb in jobs_params]}")

    with Pool(max_num_processes) as p:
        job_results = p.map(
            partial(
                execute_job,
                chunk_store=chunk_store,
                num_bytes_for_each_distance=num_bytes_for_each_distance
            ),
            jobs_params
        )

    results_matrix = [
        [
            np.zeros((
                chunk_size
                if row_index < num_chunks - 1 or num_strings % chunk_size == 0
                else num_strings % chunk_size,
                chunk_size
                if col_index < num_chunks - 1 or num_strings % chunk_size == 0
                else num_strings % chunk_size
            ), dtype=num_bytes_to_uint_dtype(num_bytes_for_each_distance))
            for col_index in range(num_chunks)
        ]
        for row_index in range(num_chunks)
    ]

    for job_result in job_results:
        for chunk_result in job_result.chunk_results:
            if isinstance(chunk_result, ChunkResultsDiagonal):
                results_matrix[chunk_result.index][chunk_result.index] = chunk_result.dist_triu
            elif isinstance(chunk_result, ChunkResultsInternal):
                results_matrix[chunk_result.row_index][chunk_result.col_index] = chunk_result.dist_matrix

    print(
        f"Block result shapes: {[[result.shape for result in results_row] for results_row in results_matrix]}")

    global_dist_triu = np.block(results_matrix)
    return global_dist_triu + global_dist_triu.T
    # np.zeros((num_strings, num_strings))


def build_seqs_distance_matrix_by_chunks(
    seqs: list[SeqRecord],
    num_chunks: int = None,
    max_num_processes: int = None,
    chunk_store: ChunkStore = dummy_chunk_store
) -> np.ndarray:
    return build_string_distance_matrix_by_chunks(
        strings=[str(seq.seq) for seq in seqs],
        num_chunks=num_chunks,
        max_num_processes=max_num_processes,
        chunk_store=chunk_store
    )



from typing import Callable
from Bio.Phylo import PhyloXML
from Bio.Phylo.PhyloXML import Sequence, Phylogeny, Property
from Bio.Phylo.BaseTree import Tree, Clade
from copy import copy


def seq_to_chrom(seq: Sequence) -> str:
    return seq.location.split(':')[0]

def sequences_from_chromosome(chromosome_label: str) -> Callable[[Sequence], bool]:
    return lambda seq : seq_to_chrom(seq) == chromosome_label

            
def filter_phylogeny(
    phylogeny: Tree,
    select_fun: Callable[[Clade], bool],
    clade_merges: dict[str, str] = None
) -> Tree:

    def filter_clade(clade: Clade) -> Clade:
        if len(clade.clades) == 0:
            if select_fun(clade):
                return clade
            else:
                return None
        new_subclades = [
            filter_clade(subclade)
            for subclade in clade.clades
        ]
        new_subclades = [
            subclade
            for subclade in new_subclades
            if subclade is not None
        ]
        if len(new_subclades) == 0:
            return None
        if len(new_subclades) == 1:
            new_subclades[0].branch_length += clade.branch_length
            if clade.name is not None:
                if (
                    new_subclades[0].name is not None and
                    len(new_subclades[0].clades) > 0
                ):
                    print(f"Clade label clash collapsing parent {clade.name} and child {new_subclades[0].name}")
                    if clade_merges is not None:
                        clade_merges[new_subclades[0].name] = clade.name
                new_subclades[0].name = clade.name
            return new_subclades[0]
        new_clade = copy(clade)
        new_clade.clades = new_subclades
        return new_clade
    
    new_root = filter_clade(phylogeny.root)
    if new_root is None:
        return None
    
    new_phylogeny = copy(phylogeny)
    new_phylogeny.root = new_root
    return new_phylogeny
            
def filter_phylogeny_by_chromosome(
    phylogeny: Tree,
    chromosome_label: str,
    clade_merges: dict[str, str] = None
) -> Tree:
    return filter_phylogeny(
        phylogeny=phylogeny,
        select_fun=lambda leaf: any([
            seq_to_chrom(seq) == chromosome_label
            for seq in leaf.sequences
        ]),
        clade_merges=clade_merges
    )
    
def split_phylogeny_by_chromosome(
    phylogeny: Tree
) -> dict[str, tuple[Tree, dict[str, str]]]:
    chromosome_labels = set([
        seq_to_chrom(seq)
        for leaf in phylogeny.get_terminals()
        for seq in leaf.sequences
    ])
    return {
        chromosome_label: filter_phylogeny_by_chromosome(
            phylogeny=phylogeny,
            chromosome_label=chromosome_label
        ) for chromosome_label in chromosome_labels
    }

def filter_hor_clade(
    hor_clade: PhyloXML.Clade,
    select_fun: Callable[[Sequence], bool],
    map_prop_fun: Callable[[Property], Property] 
) -> PhyloXML.Clade:
    filtered_seqs = [seq for seq in hor_clade.sequences if select_fun(seq)]
    if len(filtered_seqs) == 0:
        return None
    filtered_subhors = filter(
        lambda s: s is not None,
        [
            filter_hor_clade(hor_clade=subhor, select_fun=select_fun)
            for subhor in hor_clade.clades
        ]
    )
    new_hor_clade = copy(hor_clade)
    new_hor_clade.sequences = filtered_seqs
    new_hor_clade.clades = filtered_subhors
    new_hor_clade.properties = map(map_prop_fun, new_hor_clade.properties)
    return new_hor_clade
            
def filter_hor_tree(
    hor_tree: Phylogeny,
    select_fun: Callable[[Sequence], bool],
    map_prop_fun: Callable[[Property], Property] 
) -> Phylogeny:

    def filter_hor_clade(hor_clade: PhyloXML.Clade) -> PhyloXML.Clade:
        filtered_seqs = [seq for seq in hor_clade.sequences if select_fun(seq)]
        if len(filtered_seqs) == 0:
            return None
        filtered_subhors = filter(
            lambda s: s is not None,
            [
                filter_hor_clade(subhor)
                for subhor in hor_clade.clades
            ]
        )
        new_hor_clade = copy(hor_clade)
        new_hor_clade.sequences = filtered_seqs
        new_hor_clade.clades = filtered_subhors
        new_hor_clade.properties = [
            map_prop_fun(property)
            for property in new_hor_clade.properties
        ]
        return new_hor_clade

    return Phylogeny(
        name=hor_tree.name,
        root=filter_hor_clade(hor_clade=hor_tree.root)
    )
            
def filter_hor_tree_by_chromosome(
    hor_tree: Phylogeny,
    chromosome_label: str,
    map_prop_fun: Callable[[Property], Property] 
) -> Phylogeny:
    return filter_hor_tree(
        hor_tree=hor_tree,
        select_fun=lambda seq: seq_to_chrom(seq) == chromosome_label,
        map_prop_fun=map_prop_fun
    )
    
def split_hor_tree_by_chromosome(
    hor_tree: Phylogeny
) -> dict[str, Phylogeny]:
    chromosome_labels = set([
        seq_to_chrom(seq)
        for seq in hor_tree.root.sequences
    ])
    return {
        chromosome_label: filter_hor_tree_by_chromosome(
            hor_tree=hor_tree,
            chromosome_label=chromosome_label
        ) for chromosome_label in chromosome_labels
    }
    
def filter_phyloxml_by_chromosome(
    phyloxml: PhyloXML.Phyloxml,
    chromosome_label: str
) -> PhyloXML.Phyloxml:
    new_phylogenies = []
    clade_merges = {}
    
    try:
        new_phylogenies.append(filter_phylogeny_by_chromosome(
            phylogeny=phyloxml['monomers'],
            chromosome_label=chromosome_label,
            clade_merges=clade_merges
        ))
    except KeyError:
        pass
    
    def rename_families(property: Property):
        if property.ref == 'monomer_clade_seq':
            property = copy(property)
            property.value = ','.join([
                (
                    clade_merges[clade_label]
                    if clade_label in clade_merges
                    else clade_label
                )
                for clade_label in property.value.split(',')
            ]) 
        return property

    
    try:
        new_hor_tree = filter_hor_tree_by_chromosome(
            hor_tree=phyloxml['hors'],
            chromosome_label=chromosome_label,
            map_prop_fun=rename_families
        )
        new_phylogenies.append(new_hor_tree)
    except KeyError:
        pass
    
    return PhyloXML.Phyloxml(
        attributes=phyloxml.attributes,
        phylogenies=new_phylogenies,
        other=phyloxml.other
    )

def split_phyloxml_by_chromosomes(
    phyloxml: PhyloXML.Phyloxml
) -> dict[str, PhyloXML.Phyloxml]:
    chromosome_labels = set([
        seq_to_chrom(seq)
        for leaf in phyloxml['monomers'].get_terminals()
        for seq in leaf.sequences
    ])
    return {
        chromosome_label: filter_phyloxml_by_chromosome(
            phyloxml=phyloxml,
            chromosome_label=chromosome_label
        ) for chromosome_label in chromosome_labels
    }



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches
import matplotlib.colors as mcolors
import matplotlib.gridspec as gridspec
from Bio import Phylo
from .treeFromClusters import new_phylogeny

def show_hor_in_seq(
    hor_in_seq,
    reference_seq,
    ax_seq,
    seq_fg_color='black',
    seq_bg_color='#D3D3D3'
):
    bp_seq_length = len(reference_seq)

    ax_seq.set_xlim([0, bp_seq_length])
    ax_seq.get_yaxis().set_visible(False)

    # spans = hor_in_seq.spans_in_seq
    locations = hor_in_seq.locations

    # bp_start = spans[0].span_start * monomer_size
    # bp_end = (spans[-1].span_start + spans[-1].span_length) * monomer_size

    ax_seq.add_patch(patches.Rectangle(
        (0, 0), bp_seq_length, 1, facecolor=seq_bg_color))
    for location in locations:
        ax_seq.add_patch(patches.Rectangle(
            (location.start, 0 if location.strand == -1 else 0.5),
            location.end - location.start,
            0.5,
            facecolor=seq_fg_color))
    


def show_hor(
    hor_in_seq,
    hor_index,
    reference_seq,
    parent_gridspec,
    seq_fg_color='black',
    seq_bg_color='#D3D3D3',
    hor_item_size_ratio=(0.04, 0.5),
    h_space_in_hor=0.02
):
        clade_seq = hor_in_seq.hor.clade_seq
        hor_size_ratio = hor_item_size_ratio[0]*len(clade_seq)

        gs_hor = gridspec.GridSpecFromSubplotSpec(
            2, 2,
            subplot_spec=parent_gridspec[hor_index, 0],
            width_ratios=[
                hor_size_ratio, 1 - hor_size_ratio
            ],
            height_ratios=[1, 2],
            hspace=h_space_in_hor
        )

        # gs_hor = ax_hor_and_seq.add_gridspec(
        #     2, 2,
        #     width_ratios=[hor_size_ratio,1-hor_size_ratio],
        #     height_ratios=[1,1],
        #     hspace=h_space
        # )

        ax_hor = plt.subplot(gs_hor[0, 0])
        ax_seq = plt.subplot(gs_hor[1, :])

        ax_hor.set_xlim([0, len(clade_seq)])
        ax_hor.grid(True)
        ax_hor.set_xticks(np.arange(0, len(clade_seq) + 1, 1))
        ax_hor.get_yaxis().set_visible(False)
        for clade_pos, clade in enumerate(clade_seq):
            ax_hor.add_patch(patches.Rectangle(
                (clade_pos, 0), 1, 1, facecolor=clade.color.to_hex()))
            
        show_hor_in_seq(
            hor_in_seq=hor_in_seq,
            reference_seq=reference_seq,
            ax_seq=ax_seq,
            seq_fg_color=seq_fg_color,
            seq_bg_color=seq_bg_color
        )


def show_hors(hors_in_seq, reference_seq,
              tree=None,
              label='unnamed',
              color_palette=[mcolors.TABLEAU_COLORS[color_id]
                             for color_id in mcolors.TABLEAU_COLORS],
              seq_fg_color='black',
              seq_bg_color='#D3D3D3',
              tree_bg_color='#D3D3D3',
              hor_item_size_ratio=(0.04, 0.5),
              seq_height_ratio=0.5,
              fig_width=7,
              tree_height=5,
              hor_height=1,
              h_space=0.1,
              h_space_in_hor=0.02,
              label_func=(lambda clade_name: None)):

    clade_set = set(
        [clade for hor_in_seq in hors_in_seq for clade in hor_in_seq.hor.clade_seq])
    for clade_index, clade in enumerate(clade_set):
        # this is bad, but at least avoids error
        clade.color = color_palette[clade_index % 10]

    hors_height = hor_height * len(hors_in_seq)
    fig = plt.figure(constrained_layout=True, figsize=(
        fig_width, tree_height + hors_height))
    # fig.suptitle(f'HOR {label} ({bp_start}-{bp_end})')

    # print([hors_height for i in range(len(hors_in_seq))].append(tree_height))

    gs = fig.add_gridspec(
        len(hors_in_seq) + 1, 1,
        # width_ratios=[hor_size_ratio,1-hor_size_ratio],
        height_ratios=[hor_height for i in range(
            len(hors_in_seq))] + [tree_height],
        hspace=h_space
    )
    # ax_hors = [plt.subplot(gs[i, :]) for i in range(len(hors_in_seq))]
    ax_tree = plt.subplot(gs[len(hors_in_seq), :])

    for hor_index in range(len(hors_in_seq)):
        # ax_hor_and_seq = plt.subplot(gs[hor_index, 0])
        show_hor(
            hor_in_seq=hors_in_seq[hor_index],
            hor_index=hor_index,
            reference_seq=reference_seq,
            parent_gridspec=gs,
            seq_fg_color=seq_fg_color,
            seq_bg_color=seq_bg_color,
            hor_item_size_ratio=hor_item_size_ratio,
            h_space_in_hor=h_space_in_hor
        )

    ax_tree.get_yaxis().set_visible(False)
    ax_tree.get_xaxis().set_visible(False)
    # ax_tree.spines['top'].set_visible(False)
    # ax_tree.spines['right'].set_visible(False)
    # ax_tree.spines['bottom'].set_visible(False)
    # ax_tree.spines['left'].set_visible(False)
    ax_tree.get_yaxis().set_ticks([])
    # ax_tree.axis('off')
    if tree is not None:
        tree.root.color = tree_bg_color
        Phylo.draw(tree, axes=ax_tree, label_func=label_func)
        # Phylo.draw_graphviz(tree, axes=ax_tree)

    for clade in clade_set:
        clade.color = None
    if tree is not None:
        tree.root.color = None


def show_hor_tree(hor_tree_root, reference_seq, tree, path=[], level=0):
    if not hor_tree_root.sub_hors:
        return
    print(f'Subtree: {path}')
    show_hors(hor_tree_root.sub_hors, reference_seq, tree=tree)
    for branch_index, sub_hor in enumerate(hor_tree_root.sub_hors):
        show_hor_tree(sub_hor, reference_seq, tree=new_phylogeny(
            tree.common_ancestor(sub_hor.hor.clade_seq)), path=path + [branch_index + 1])



from dataclasses import dataclass
from Bio.Phylo.BaseTree import Clade, Tree

@dataclass
class CladeSortResult:
    sorted_clade: Clade
    min_label: str    

def sort_clade_by_leaf_names(clade: Clade) -> CladeSortResult:
    if len(clade.clades) == 0:
        return CladeSortResult(sorted_clade=clade, min_label=clade.name)
    sorted_subclade_results = [
        sort_clade_by_leaf_names(subclade) for subclade in clade.clades
    ]
    sorted_subclade_results.sort(
        key=lambda c: c.min_label
    )
    clade.clades = [
        clade_sort_result.sorted_clade
        for clade_sort_result in sorted_subclade_results
    ]
    return CladeSortResult(
        sorted_clade=clade,
        min_label=sorted_subclade_results[0].min_label
    )
    
def sort_phylogeny_by_leaf_names(phylogeny: Tree):
    sort_clade_by_leaf_names(phylogeny.root)


from collections.abc import Iterable
from typing_extensions import Self
from .loops import LoopSpanInSeq, LoopInSeq
from Bio.Phylo import BaseTree
from Bio.Phylo.PhyloXML import Clade, Phylogeny, Sequence, Property
from Bio.SeqFeature import SeqFeature, SimpleLocation, CompoundLocation, Location
from typing import Optional

class HOR:
    clade_seq: list[BaseTree.Clade]

    def __init__(self, clade_seq: list[BaseTree.Clade]):
        self.clade_seq = clade_seq

    def __str__(self):
        return ''.join([clade.id if clade.id is not None else '*' for clade in self.clade_seq])

class HORInSeq:
    hor: HOR
    locations: list[Location]
    spans_in_seq: list[LoopSpanInSeq]
    super_hor: Self
    sub_hors: list[Self]

    def __init__(
        self,
        hor: HOR,
        spans_in_seq: list[LoopSpanInSeq] = [],
        locations: list[Location] = None
    ):
        self.hor = hor
        self.spans_in_seq = spans_in_seq
        self.locations = locations

    def add_span(self, span_in_seq: LoopSpanInSeq):
        self.spans_in_seq.append(span_in_seq)

    def __str__(self):
        return (
            f'{self.loop}' +
            (
                f' in {",".join([str(span) for span in self.spans_in_seq])}'
                    if len(self.spans_in_seq) > 0 else ''
            )
        )

def seq_span_to_location(
    span: LoopSpanInSeq,
    seq_locations: list[SimpleLocation]
) -> SimpleLocation:
    start_location = seq_locations[span.span_start]
    end_location = seq_locations[span.span_start + span.span_length - 1]
    ref = start_location.ref
    strand = start_location.strand
    return SimpleLocation(
        ref=ref,
        strand=strand,
        start=min(start_location.start,end_location.start),
        end=max(end_location.end,start_location.end)
    )

def seq_spans_to_compound_location(
    spans_in_seq: Iterable[LoopSpanInSeq],
    seq_locations: list[SimpleLocation]
) -> CompoundLocation:
    return CompoundLocation([seq_span_to_location(span, seq_locations) for span in spans_in_seq])

def loop_to_HOR(
    loop_in_seq: LoopInSeq,
    clades: list[Clade],
    seq_locations: list[SimpleLocation]=None
) -> HORInSeq:
    hor = HOR([clades[clade_index] for clade_index in loop_in_seq.loop.loop_seq])
    return HORInSeq(
        hor,
        spans_in_seq=loop_in_seq.spans_in_seq,
        locations=(
            [seq_span_to_location(span, seq_locations) for span in loop_in_seq.spans_in_seq]
            if seq_locations is not None else None
        )
    )

def loops_to_HORs(
    loops_in_seq: Iterable[LoopInSeq], clades: list[Clade], seq_locations=None
) -> list[HORInSeq]:
    return [loop_to_HOR(loop_in_seq, clades, seq_locations=seq_locations) for loop_in_seq in loops_in_seq]

def name_hor_tree(
    hor: HORInSeq, 
    node_prefix: str = '',
    clade_name_prefix: str = 'F',
    hor_name_prefix: str = 'H',
    level_separator: str = '_'
) -> None:
    hor.id = f'{hor_name_prefix}{node_prefix}'
    hor.feature = SeqFeature(
        id=hor.id,
        location=(
            None
                if hor.locations is None or len(hor.locations) == 0
            else hor.locations[0]
                if len(hor.locations) == 1
            else CompoundLocation(hor.locations)
        )
    )
    clade_count = 0
    if len(hor.hor.clade_seq) == 1:
        clade = hor.hor.clade_seq[0]
        if clade.name is None:
            clade.name = f'{clade_name_prefix}{node_prefix}'
    for clade in hor.hor.clade_seq:
        if clade.name is None:
            clade_count += 1
            clade.name = f'{clade_name_prefix}{node_prefix}#{clade_count}'
    common_prefix_for_sub_hors = f"{node_prefix}{level_separator if len(node_prefix) > 0 else ''}"
    if hasattr(hor, 'sub_hors'):
        for sub_hor_index, sub_hor in enumerate(hor.sub_hors):
            name_hor_tree(
                sub_hor,
                node_prefix=f'{common_prefix_for_sub_hors}{sub_hor_index + 1}')
        
def hor_tree_as_phyloxml_phylogeny(
    hor_tree_root: HORInSeq,
    name: str = 'hors',
    set_branch_lengths: bool = True,
    clade_depths: Optional[dict[Clade, float]] = None
) -> Phylogeny:
    if set_branch_lengths and clade_depths is None:
        clade_depths = hor_tree_root.hor.clade_seq[0].root.depths()

    def hor_to_clade(hor: HORInSeq, parent_hor_depth: float = 0) -> Clade:
        clade_seq_str = ",".join([clade.name for clade in hor.hor.clade_seq])
        hor_depth = (
            max([clade_depths[clade] for clade in hor.hor.clade_seq])
            if set_branch_lengths else parent_hor_depth + 1
        )
        return Clade(
            # node_id=hor.id,
            name=hor.id,
            sequences=[Sequence(type='dna', location=location) for location in hor.locations],
            branch_length=hor_depth - parent_hor_depth,
            clades=
                [hor_to_clade(sub_hor, parent_hor_depth=hor_depth) for sub_hor in hor.sub_hors]
                if hasattr(hor, 'sub_hors') else [],
            properties=[Property(value=clade_seq_str, ref='monomer_clade_seq', applies_to='clade', datatype='xsd:string')]
        )

    return Phylogeny(root=hor_to_clade(hor_tree_root), name=name)




from Bio.Phylo.BaseTree import Tree, Clade
from Bio.SeqFeature import SimpleLocation

from .loops import LoopInSeq, LoopSpanInSeq
from .phylogeny_to_levels import parse_location

def locationContainsLocation(
    containingLocation: SimpleLocation,
    containedLocation: SimpleLocation
) -> bool:
    return (
        containedLocation.start >= containingLocation.start and 
        containedLocation.end <= containingLocation.end
    )

def locationListContainsLocation(
    containingLocationList: list[SimpleLocation],
    containedLocation: SimpleLocation
) -> bool:
    for containingLocation in containingLocationList:
        if locationContainsLocation(
            containingLocation=containingLocation,
            containedLocation=containedLocation
        ):
            return True
    return False

def checkLocationListContainsLocationList(
    containingLocationList: list[SimpleLocation],
    containedLocationList: list[SimpleLocation],
    containementNotFoundTemplate: str = "HOR location {containedLocation} not contained in parent HOR locations {containingLocationList}"
):
    for containedLocation in containedLocationList:
        if not locationListContainsLocation(
            containingLocationList=containingLocationList,
            containedLocation=containedLocation
        ):
            print(containementNotFoundTemplate.format(
                containingLocationList=containingLocationList,
                containedLocation=containedLocation
            ))
            
def locationOverlapsLocation(
    location1: SimpleLocation,
    location2: SimpleLocation
) -> bool:
    return (
        location1.start < location2.end and 
        location1.end > location2.start
    )

def locationListOverlapsLocationList(
    locationList1: list[SimpleLocation],
    locationList2: list[SimpleLocation]
) -> bool:
    for location1 in locationList1:
        for location2 in locationList2:
            if locationOverlapsLocation(
                location1=location1,
                location2=location2
            ):
                return True
    return False

def checkLocationListOverlapsLocationList(
    locationList1: list[SimpleLocation],
    locationList2: list[SimpleLocation],
    foundOverlapTemplate: str = "Found overlap between {location1} and {location2}"
):
    for location1 in locationList1:
        for location2 in locationList2:
            if locationOverlapsLocation(
                location1=location1,
                location2=location2
            ):
                print(foundOverlapTemplate.format(
                    location1=location1,
                    location2=location2
                ))

def checkLocationListSelfOverlap(
    locationList: list[SimpleLocation],
    foundOverlapTemplate: str = "Found overlap between {location1} and {location2}"
):
    controlledLocationList = []
    for location in locationList:
        checkLocationListOverlapsLocationList(
            locationList1=controlledLocationList,
            locationList2=[location],
            foundOverlapTemplate=foundOverlapTemplate
        )
        controlledLocationList.append(location)

def checkHORCoherence(horClade: Clade, path: list[int]):
    parentLocations: list[SimpleLocation] = [parse_location(seq.location) for seq in horClade.sequences]
    horName = horClade.name if horClade.name is not None else 'unknown'
    checkLocationListSelfOverlap(
        parentLocations, 
        foundOverlapTemplate=(
            f"In clade {horName} (path: {path}) " +
            "found self overlap between {location1} and {location2}"
        )
    )
    locationsFound: list[SimpleLocation] = []
    for subhor in horClade.clades:
        subhorName = subhor.name if subhor.name is not None else 'unknown'
        newLocations = [parse_location(seq.location) for seq in subhor.sequences]
        checkLocationListContainsLocationList(
            containingLocationList=parentLocations,
            containedLocationList=newLocations,
            containementNotFoundTemplate=(
                f"Location of HOR {subhorName} " +
                "({containedLocation}) " +
                f"not found among locations of HOR {horName} " +
                "({containingLocationList})"
            )
        )
        checkLocationListOverlapsLocationList(
            locationList1=locationsFound,
            locationList2=newLocations,
            foundOverlapTemplate=(
                f"In clade {subhorName} " +
                "found sibling overlap between {location1} and {location2}"
            )
        )
        locationsFound.extend(newLocations)
    for subhorIndex, subhor in enumerate(horClade.clades):
        checkHORCoherence(subhor, path=path + [subhorIndex])
        
def checkHORTreeCoherence(horTree: Tree):
    checkHORCoherence(horClade=horTree.root, path=[])
    
    
def loopSpanInSeqOverlapsLoopSpanInSeq(
    loopSpanInSeq1: LoopSpanInSeq,
    loopSpanInSeq2: LoopSpanInSeq
) -> bool:
    return (
        loopSpanInSeq1.span_start < loopSpanInSeq2.span_start + loopSpanInSeq2.span_length and 
        loopSpanInSeq1.span_start + loopSpanInSeq1.span_length > loopSpanInSeq2.span_start
    )
         
def checkSpanInSeqListOverlapsSpanInSeqList(
    loopSpanInSeqList1: list[LoopSpanInSeq],
    loopSpanInSeqList2: list[LoopSpanInSeq],
    foundOverlapTemplate: str = "Found overlap between {loopSpanInSeq1} and {loopSpanInSeq2}"
):
    for loopSpanInSeq1 in loopSpanInSeqList1:
        for loopSpanInSeq2 in loopSpanInSeqList2:
            if loopSpanInSeqOverlapsLoopSpanInSeq(
                loopSpanInSeq1=loopSpanInSeq1,
                loopSpanInSeq2=loopSpanInSeq2
            ):
                print(foundOverlapTemplate.format(
                    loopSpanInSeq1=loopSpanInSeq1,
                    loopSpanInSeq2=loopSpanInSeq2
                ))

def checkLoopInSeqSelfOverlap(
    loopInSeq: LoopInSeq,
    foundOverlapTemplate: str = "Found overlap between {loopSpanInSeq1} and {loopSpanInSeq2}"
):
    controlledSpanInSeqList = []
    for spanInSeq in loopInSeq.spans_in_seq:
        checkSpanInSeqListOverlapsSpanInSeqList(
            loopSpanInSeqList1=controlledSpanInSeqList,
            loopSpanInSeqList2=[spanInSeq],
            foundOverlapTemplate=foundOverlapTemplate
        )
        controlledSpanInSeqList.append(spanInSeq)
            
# def checkLoopInSeqOverlapsLoopInSeq(
#     loopInSeq1: LoopInSeq,
#     loopInSeq2: LoopInSeq
# ) -> bool:
#     return (
#         location1.start < location2.end and 
#         location1.end > location2.start
#     )


def spanOperlapsSpan(span1: tuple[int,int], span2: tuple[int,int]):
    return (
        span1[0] < span2[1] and 
        span1[1] > span2[0]
    )

def checkSpanListOverlapsSpanList(
    spanList1: list[tuple[int,int]],
    spanList2: list[tuple[int,int]],
    foundOverlapTemplate: str = "Found overlap between {span1} and {span2}"
):
    for span1 in spanList1:
        for span2 in spanList2:
            if spanOperlapsSpan(
                span1=span1,
                span2=span2
            ):
                print(foundOverlapTemplate.format(
                    span1=span1,
                    span2=span2
                ))

def checkSpanListSelfOverlap(
    spanList: list[tuple[int,int]],
    foundOverlapTemplate: str = "Found overlap between {loopSpanInSeq1} and {loopSpanInSeq2}"
):
    controlledSpanList = []
    for span in spanList:
        checkSpanListOverlapsSpanList(
            spanList1=controlledSpanList,
            spanList2=[span],
            foundOverlapTemplate=foundOverlapTemplate
        )
        controlledSpanList.append(span)



from itertools import accumulate
from .featureUtils import get_seq_as_txt

def normalize_loop(loop_seq: list[int]) -> tuple[list[int], int]:
    def invert_pos(pos: int) -> int:
        return len(loop_seq) - pos if pos > 0 else 0
    options = []
    for start_index in range(len(loop_seq)):
        options.append(loop_seq[start_index:] + loop_seq[:start_index] + [start_index])
    options.sort()
    return (options[0][:-1],invert_pos(options[0][-1]))

def denormalize_loop(
    loop_seq: list[int],
    in_loop_start: int
) -> list[int]:
    return loop_seq[in_loop_start:] + loop_seq[:in_loop_start]

class Loop:
    loop_seq: list[int]

    def __init__(self, loop_seq: list[int]):
        self.loop_seq = loop_seq

    def __str__(self):
        return get_seq_as_txt(self.loop_seq)

class LoopSpanInSeq:
    span_start: int
    span_length: int
    num_of_laps: int
    in_loop_start: int
    num_mismatches: int

    def __init__(
        self,
        span_start: int, span_length: int,
        num_of_laps: int, in_loop_start: int,
        num_mismatches: int = 0
    ):
        if span_length <= 0:
            raise Exception(f"Span length {span_length} not permitted")
        self.span_start = span_start
        self.span_length = span_length
        self.num_of_laps = num_of_laps
        self.in_loop_start = in_loop_start
        self.num_mismatches = num_mismatches

    def __str__(self):
        return (
            f'[{self.span_start}:{self.span_start + self.span_length}]' +
            (f'#{self.in_loop_start}' if self.in_loop_start != 0 else '') +
            (f'mm({self.num_mismatches})' if self.num_mismatches != 0 else '')
        )

class LoopInSeq:
    loop: Loop
    spans_in_seq: list[LoopSpanInSeq]

    def __init__(self, loop: Loop, spans_in_seq: list[LoopSpanInSeq] = []):
        self.loop = loop
        self.spans_in_seq = spans_in_seq

    def add_span(self, span_in_seq: LoopSpanInSeq):
        self.spans_in_seq.append(span_in_seq)

    def __str__(self):
        return (
            f'{self.loop}' +
            (
                f' in {",".join([str(span) for span in self.spans_in_seq])}'
                    if len(self.spans_in_seq) > 0 else ''
            )
        )
        
def loop_to_spans(loop_in_seq: LoopInSeq) -> list[tuple[int,int]]:
    spans: list[tuple[int,int]] = []
    for loop_span in loop_in_seq.spans_in_seq:
        loop_span_end = loop_span.span_start + loop_span.span_length
        if len(spans) > 0 and loop_span.span_start <= spans[-1][1]:
            if loop_span_end > spans[-1][1]:
                spans.append((spans[-1][1], loop_span_end))
        else:
            spans.append((loop_span.span_start, loop_span_end))
    return spans

def extract_spans_in_seqs(
    spans_in_seq: list[LoopSpanInSeq],
    spans_to_extract: list[tuple[int,int]],
    loop_size: int
) -> list[LoopSpanInSeq]:
    new_spans_in_seq: list[LoopSpanInSeq] = []
    span_to_extract_index = 0
    for span_in_seq in spans_in_seq:
        span_in_seq_end = span_in_seq.span_start + span_in_seq.span_length
        while span_to_extract_index < len(spans_to_extract):
            span_to_extract_start, span_to_extract_end = spans_to_extract[span_to_extract_index]
            if (span_to_extract_start < span_in_seq_end and span_to_extract_end > span_in_seq.span_start):
                new_span_start = max(span_in_seq.span_start, span_to_extract_start)
                new_span_end = min(span_in_seq_end, span_to_extract_end)
                new_span_length = new_span_end - new_span_start
                new_in_loop_start = (
                    (span_in_seq.in_loop_start + new_span_start - span_in_seq.span_start) % loop_size
                )
                new_spans_in_seq.append(LoopSpanInSeq(
                    span_start=new_span_start,
                    span_length=new_span_length,
                    num_of_laps=new_span_length // loop_size,
                    in_loop_start=new_in_loop_start,
                    num_mismatches=span_in_seq.num_mismatches
                ))
            if span_to_extract_start < span_in_seq.span_start:
                spans_to_extract.insert(
                    span_to_extract_index,
                    (
                        span_to_extract_start,
                        min(span_to_extract_end, span_in_seq.span_start)
                    )
                )
                span_to_extract_index += 1
            if (span_to_extract_end > span_in_seq_end):
                spans_to_extract[span_to_extract_index] = (
                    max(span_in_seq_end, span_to_extract_start),
                    span_to_extract_end
                 )
                break
            else:
                del spans_to_extract[span_to_extract_index]
    return new_spans_in_seq

def intersect_spans(
    spans_a: list[tuple[int,int]],
    spans_b: list[tuple[int,int]]
) -> list[tuple[int,int]]:
    intersection_spans = []
    next_span_b_index = 0
    for start_pos_a, end_pos_a in spans_a:
        for span_b_index_offset, span_b_boundaries in enumerate(spans_b[next_span_b_index:]):
            start_pos_b, end_pos_b = span_b_boundaries
            if (start_pos_b < end_pos_a and end_pos_b > start_pos_a):
                new_span_start = max(start_pos_a, start_pos_b)
                new_span_end = min(end_pos_a, end_pos_b)
                intersection_spans.append((new_span_start, new_span_end))
            if (end_pos_b > end_pos_a):
                next_span_b_index += span_b_index_offset
                break
    return intersection_spans

def complement_of_spans(
    spans: list[tuple[int,int]],
    whole_range: int
) -> list[tuple[int,int]]:
    if len(spans) == 0:
        return [(0, whole_range)]
    complement = []
    if spans[0][0] > 0:
        complement.append((0, spans[0][0]))
    complement.extend([
        (spans[i][1], spans[i+1][0])
        for i in range(len(spans) - 1)
    ])
    if spans[-1][1] < whole_range:
        complement.append((spans[-1][1], whole_range))
    return complement
    

def spans_difference(
    spans_a: list[tuple[int,int]],
    spans_b: list[tuple[int,int]]
):
    if len(spans_a) == 0:
        return []
    return intersect_spans(
        spans_a,
        complement_of_spans(spans_b, whole_range=spans_a[-1][1]))
        

def find_loops(
    seqs: list[list[int]] = None,
    whole_seq: list[int] = None,
    gap_indices: list[int] = None,
    seq_spans: list[tuple[int,int]] = None,
    min_loop_size: int = 2, max_loop_size: int = 30, min_loops: int = 3,
    min_diversity: int = 1,
    allowed_mismatch_rate: float = 0,
    allow_overlap: bool = False
) -> list[LoopInSeq]:
    
    if seqs is None:
        if seq_spans is None:
            if gap_indices is None:
                gap_indices = []
            seq_offsets = [0] + gap_indices + [len(whole_seq)]
            seq_spans = [
                (seq_offsets[i],seq_offsets[i+1])
                for i in range(len(gap_indices) + 1)
            ]
            seqs = [
                whole_seq[seq_offsets[i]:seq_offsets[i+1]]
                for i in range(len(gap_indices) + 1)
            ]
        else:
            seq_offsets = [seq_start for seq_start, seq_end in seq_spans]
            seqs = [
                whole_seq[seq_start:seq_end]
                for seq_start, seq_end in seq_spans
            ]
    else:
        seq_offsets = [0] + list(accumulate([len(seq) for seq in seqs]))
        seq_spans = [
            (seq_offsets[i],seq_offsets[i+1])
            for i in range(len(seqs))
        ]
        
    loops_found: dict[str,LoopInSeq] = {} #defaultdict(list)
    for seqIndex, seq in enumerate(seqs):

        curr_loops = [0] * (max_loop_size - min_loop_size + 1)
        curr_loops_mismatches = [0] * (max_loop_size - min_loop_size + 1)

        def last_of_size(curr_position, loop_size):
            loop_size_offset = loop_size - min_loop_size
            if curr_loops[loop_size_offset] >= (min_loops - 1) * loop_size:
                loop_start = curr_position - curr_loops[loop_size_offset] - loop_size
                loop_length = curr_loops[loop_size_offset] + loop_size
                loop_laps = loop_length // loop_size
                loop_items = seq[loop_start:loop_start + loop_size]
                if len(set(loop_items)) >= min_diversity:
                    normal_loop, in_loop_start_position = normalize_loop(loop_items)
                    normal_loop_str = str(normal_loop)
                    loop_span = LoopSpanInSeq(
                        seq_offsets[seqIndex] + loop_start,
                        loop_length, loop_laps, in_loop_start_position,
                        num_mismatches=curr_loops_mismatches[loop_size_offset])
                    if normal_loop_str not in loops_found:
                        loops_found[normal_loop_str] = LoopInSeq(
                            Loop(normal_loop),
                            [loop_span]
                        )
                    else:
                        loops_found[normal_loop_str].add_span(loop_span)
        
        for curr_position, curr_symbol in enumerate(seq):
            max_loop_length_closed = 0
            for loop_size in range(min_loop_size, min(max_loop_size, curr_position) + 1):
                loop_size_offset = loop_size - min_loop_size
                if seq[curr_position - loop_size] == curr_symbol:
                    curr_loops[loop_size_offset] += 1
                elif curr_loops_mismatches[loop_size_offset] + 1 <= (curr_loops[loop_size_offset] + 1) * allowed_mismatch_rate:
                    curr_loops[loop_size_offset] += 1
                    curr_loops_mismatches[loop_size_offset] += 1
                else:
                    if curr_loops[loop_size_offset] > max_loop_length_closed:
                        last_of_size(curr_position, loop_size)
                        max_loop_length_closed = curr_loops[loop_size_offset]
                    curr_loops[loop_size_offset] = 0
                    curr_loops_mismatches[loop_size_offset] = 0
        
        max_loop_length_closed = 0
        for loop_size in range(min_loop_size, max_loop_size + 1):
            loop_size_offset = loop_size - min_loop_size
            if curr_loops[loop_size_offset] > max_loop_length_closed:
                last_of_size(len(seq), loop_size)
                max_loop_length_closed = curr_loops[loop_size_offset]

    loops = list(loops_found.values())

    if not allow_overlap:
        loops.sort(key=lambda loop: len(loop.loop.loop_seq), reverse=True)
        filtered_loops = []
        available_spans = seq_spans
        for loop in loops:
            extracted_spans = extract_spans_in_seqs(
                spans_in_seq=loop.spans_in_seq,
                spans_to_extract=available_spans,
                loop_size=len(loop.loop.loop_seq)
            )
            extracted_spans = [span for span in extracted_spans if span.num_of_laps >= min_loops]
            if len(extracted_spans) > 0:
                filtered_loop = LoopInSeq(
                    loop=loop.loop,
                    spans_in_seq=extracted_spans
                )
                filtered_loops.append(filtered_loop)
        loops = filtered_loops

    for loop in loops:
        spans = loop.spans_in_seq
        if all([span.in_loop_start == spans[0].in_loop_start for span in spans]):
            loop.loop.loop_seq = denormalize_loop(loop.loop.loop_seq, spans[0].in_loop_start)
            for span in spans:
                span.in_loop_start = 0

    return loops



from Bio.SeqRecord import SeqRecord
import editdistance
import numpy as np


def build_string_distance_matrix(strings: list[str]) -> np.ndarray:
    seq_triu = np.array([
            (0 if j <= i else editdistance.eval(seq_i, strings[j]))
            for i, seq_i in enumerate(strings)
            for j in range(len(strings))
    ])
    seq_triu.shape = (len(strings), len(strings))
    return seq_triu + seq_triu.T


def build_seqs_distance_matrix(seqs: list[SeqRecord]) -> np.ndarray:
    return build_string_distance_matrix([str(seq.seq) for seq in seqs])


def min_distance(distance_matrix: np.ndarray) -> any:
    return min(distance_matrix[np.triu_indices(distance_matrix.shape[0],1)])


def distance_values(matrix: np.ndarray):
    return matrix[np.triu_indices(matrix.shape[0], k = 1)]


from dataclasses import dataclass
import json
from typing import Optional
import numpy as np
from sklearn.cluster import AgglomerativeClustering
from Bio.SeqFeature import SeqFeature
from Bio.SeqRecord import SeqRecord
from Bio.Phylo.PhyloXML import Clade, Phylogeny
from .treeFromClusters import feature_to_leave, new_phylogeny
from .featureUtils import feature_to_seq, label_to_feature
from .assertions import assert_equal

class SimplePhylogeny:
    num_leaves: int
    children: list[list[int]]
    root_clade_index: int
    num_clades: int
    
    def __init__(
        self,
        num_leaves: int,
        children: list[list[int]]
    ):
        self.num_leaves = num_leaves
        self.children = children
        self.num_clades = num_leaves + len(children)
        self.root_clade_index = self.num_clades - 1
    
    def get_clade_children(self, clade_index: int) -> list[int]:
        return (
            [] if clade_index < self.num_leaves
            else self.children[clade_index - self.num_leaves]
        )
    
    # def set_clade_children(self, clade_index: int, children: list[int]):
    #     return (
    #         [] if clade_index < self.num_leaves
    #         else self.children[clade_index - self.num_leaves]
    #     )
    
    def get_clade_height(self, clade_index: int) -> int:
        return (
            0 if clade_index < self.num_leaves
            else 1 + max([
                self.get_clade_height(subclade_index)
                for subclade_index in self.get_clade_children(clade_index)
            ])
        )
        
    def get_leaves_in_order(
        self, clade_index: Optional[int] = None
    ) -> list[int]:
        if clade_index is None:
            clade_index = self.root_clade_index
        subclades = self.get_clade_children(clade_index)
        if len(subclades) == 0:
            return [clade_index]
        else:
            return [
                leave
                for subclade in subclades
                for leave in self.get_leaves_in_order(subclade)
            ]

class SimplePhylogenyWithDistances(SimplePhylogeny):
    max_distances: list[int]
    
    def __init__(
        self,
        num_leaves: int,
        children: list[list[int]],
        max_distances: list[int]
    ):
        super().__init__(num_leaves=num_leaves, children=children)
        if max_distances is None:
            max_distances = [
                self.get_clade_height(num_leaves + internal_clade_index)
                for internal_clade_index in enumerate(children)
            ]
        self.max_distances = max_distances
    
    def get_clade_distance(self, clade_index: int):
        return (
            0 if clade_index < self.num_leaves
            else (self.max_distances[clade_index - self.num_leaves])
        )
    
class SimplePhylogenyWithBranchLengths(SimplePhylogeny):
    branch_lengths: Optional[list[int]] = None
    
    def __init__(
        self,
        num_leaves: int,
        children: list[list[int]],
        branch_lengths: list[int]
    ):
        super().__init__(num_leaves=num_leaves, children=children)
        self.branch_lengths = branch_lengths
    
    def get_branch_length(self, clade_index: int):
        return self.branch_lengths[clade_index]
    
def save_phylogeny_with_branch_lengths(
    phylogeny: SimplePhylogenyWithBranchLengths,
    filename: str
):
    with open(filename + '.json', 'w') as f:
        json.dump({
            'num_leaves': phylogeny.num_leaves,
            'children': phylogeny.children,
            'branch_lengths': phylogeny.branch_lengths
        }, f)
        
def save_phylogeny_with_distances(
    phylogeny: SimplePhylogenyWithDistances,
    filename: str
):
    with open(filename + '.json', 'w') as f:
        json.dump({
            'num_leaves': phylogeny.num_leaves,
            'children': phylogeny.children,
            'max_distances': phylogeny.max_distances
        }, f)
        
def load_phylogeny_with_branch_lengths(
    filename: str
) -> SimplePhylogenyWithBranchLengths:
    with open(filename + '.json') as f:
        dict = json.load(f)
        return SimplePhylogenyWithBranchLengths(
            num_leaves=dict['num_leaves'],
            children=dict['children'],
            branch_lengths=dict['branch_lengths']
        )
        
def load_phylogeny_with_distances(
    filename: str
) -> SimplePhylogenyWithDistances:
    with open(filename + '.json') as f:
        dict = json.load(f)
        return SimplePhylogenyWithDistances(
            num_leaves=dict['num_leaves'],
            children=dict['children'],
            max_distances=dict['max_distances']
        )
        
def compact_phylogeny(
    input_phylogeny: SimplePhylogenyWithDistances
) -> SimplePhylogenyWithDistances:
    def clades_to_descendants_at_distance(clade_indeces: list[int], distance: int) -> list[int]:
        return [
            descendant_clade
            for child_clade_index in clade_indeces
            for descendant_clade in clade_to_descendants_at_distance(child_clade_index, distance)
        ]

    def clade_to_descendants_at_distance(clade_index: int, distance: int) -> list[int]:
        if input_phylogeny.get_clade_distance(clade_index) < distance:
            return [clade_index]
        descendant_clades_at_distance = clades_to_descendants_at_distance(
            input_phylogeny.get_clade_children(clade_index), distance
        )
        if len(descendant_clades_at_distance) == 0:
            return [clade_index]
        return descendant_clades_at_distance
    
    new_internal_clades_distances = []
    new_internal_clades_children = []
    
    def compact_clade(clade_index: int) -> int:
        subclades = input_phylogeny.get_clade_children(clade_index)
        max_distance = input_phylogeny.get_clade_distance(clade_index)
        if len(subclades) == 0:
            return clade_index
        descendants = clades_to_descendants_at_distance(
            subclades,
            max_distance
        )
        new_internal_clades_children.append([compact_clade(clade) for clade in descendants])
        new_internal_clades_distances.append(max_distance)
        new_clade_index = input_phylogeny.num_leaves + len(new_internal_clades_distances) - 1
        return new_clade_index 

    new_root_clade_index = compact_clade(input_phylogeny.root_clade_index)
    
    return SimplePhylogenyWithDistances(
        num_leaves=input_phylogeny.num_leaves,
        children=new_internal_clades_children,
        max_distances=new_internal_clades_distances
    )
    
def distances_to_branch_lengths(
    input_phylogeny: SimplePhylogenyWithDistances
) -> SimplePhylogenyWithBranchLengths:
    branch_lengths = [0] * (input_phylogeny.num_clades)
    def set_clade_branch_length(clade_index: int):
        for subclade_index in input_phylogeny.get_clade_children(clade_index):
            branch_lengths[subclade_index] = (
                input_phylogeny.get_clade_distance(clade_index) -
                input_phylogeny.get_clade_distance(subclade_index)
            ) / 2
            set_clade_branch_length(subclade_index)
    
    set_clade_branch_length(input_phylogeny.root_clade_index)
    
    return SimplePhylogenyWithBranchLengths(
        num_leaves=input_phylogeny.num_leaves,
        children=input_phylogeny.children,
        branch_lengths=branch_lengths
    )
    

def build_phylogeny(
    simple_phylogeny: SimplePhylogenyWithBranchLengths,
    items_as_seq_features: list[SeqFeature]
) -> Phylogeny:
    
    def build_clade(clade_index: int):
        subclade_indices = simple_phylogeny.get_clade_children(clade_index)
        branch_length = simple_phylogeny.get_branch_length(clade_index)
        if len(subclade_indices) == 0:
            return feature_to_leave(items_as_seq_features[clade_index], branch_length=branch_length)
        subclades = [build_clade(subclade_index) for subclade_index in subclade_indices]
        return Clade(clades=subclades, branch_length=branch_length)
        
    root_clade = build_clade(simple_phylogeny.root_clade_index)
    phylogeny = new_phylogeny(root_clade)

    return phylogeny

@dataclass
class CladeSortResult:
    subclade_index: int
    min_index: int    

def sort_by_leaf_indexes(phylogeny: SimplePhylogeny):
    
    def sort_clade_by_leaf_indices(clade_index: int):
        if len(phylogeny.get_clade_children(clade_index)) == 0:
            return clade_index
        subclade_results = [
            CladeSortResult(
                subclade_index,
                sort_clade_by_leaf_indices(subclade_index)
            )
            for subclade_index in phylogeny.get_clade_children(clade_index)
        ]
        subclade_results.sort(
            key=lambda c: c.min_index
        )
        phylogeny.children[clade_index - phylogeny.num_leaves] = [
            subclade_result.subclade_index
            for subclade_result in subclade_results
        ]
        return subclade_results[0].min_index
    
    sort_clade_by_leaf_indices(phylogeny.root_clade_index)
        
def get_clades_by_level(
    phylogeny: SimplePhylogeny
) -> tuple[list[list[int]],list[list[int]]]:
    
    clades_by_level = []
    children_num_by_level = []

    def set_by_height(clade_height: int, clade_index: int, children_num: int):
        if clade_height >= len(clades_by_level):
            missing_levels = clade_height - len(clades_by_level) + 1
            clades_by_level.extend([[]] * missing_levels)
            children_num_by_level.extend([[]] * missing_levels)
        clades_by_level[clade_height].append(clade_index)
        children_num_by_level[clade_height].append(children_num)

    def extract_clades_by_height(clade_index: int):
        subclade_indices = phylogeny.get_clade_children(clade_index)
        subclade_heights = [
            phylogeny.get_clade_height(subclade_index)
            for subclade_index in subclade_indices
        ]
        clade_height = (
            0 if len(subclade_indices) == 0
            else 1 + max(subclade_heights)
        )
        for subclade_position, subclade_height in enumerate(subclade_heights):
            subclade_index = subclade_indices[subclade_position]
            extract_clades_by_height(subclade_index)
            for height in range(subclade_height + 1, clade_height):
                set_by_height(clade_height=height, clade_index=subclade_indices[subclade_position], children_num=1)            
        set_by_height(clade_height=clade_height, clade_index=clade_index, children_num=len(subclade_indices))

    extract_clades_by_height(phylogeny.root_clade_index)
    return clades_by_level, children_num_by_level

@dataclass
class ClusteringToPhylogenyResult:
    phylogeny: Phylogeny
    item_position_to_leaf_index: list[int]

def clustering_to_phylogeny(
    clustering: Optional[AgglomerativeClustering] = None,
    item_vs_position_array = None,
    items_as_seq_records: Optional[list[SeqRecord]] = None,
    items_as_seq_features: Optional[list[SeqFeature]] = None,
    seq_references: Optional[dict[SeqRecord]] = None,
    dist_matrix: Optional[np.ndarray] = None,
    compute_distances: bool = True,
    linkage : str = 'single',
    metric : str = 'euclidean',
    sort: bool = True
) -> ClusteringToPhylogenyResult:
    
    assert_equal([
        'clustering.n_leaves_',
        'len(item_vs_position_array)',
        'len(items_as_seq_records)',
        'len(items_as_seq_features)',
        'dist_matrix.shape[0]',
        'dist_matrix.shape[1]'
    ], locals=locals())
    
    if clustering is None:
        clustering = AgglomerativeClustering(
            metric=metric if dist_matrix is None else 'precomputed',
            compute_full_tree=True,
            linkage=linkage,
            compute_distances = compute_distances,
            n_clusters=1)


    if (items_as_seq_features is not None and
        seq_references is not None and
        items_as_seq_records is None
    ):
        items_as_seq_records = [
            feature_to_seq(seq_feature, references=seq_references)
            for seq_feature in items_as_seq_features
        ]
    
    if items_as_seq_features is None and items_as_seq_records is not None:
        items_as_seq_features = [
            label_to_feature(seq_record.id)
            for seq_record in items_as_seq_records
        ]
    
    if items_as_seq_features is None:
        raise Exception("No information on sequence positions")

    if not hasattr(clustering, 'children_'):
        
        if dist_matrix is not None:
            clustering.fit(dist_matrix)
        else:
            if (items_as_seq_records is not None and 
                item_vs_position_array is None
            ):
                item_vs_position_array = [
                    str(seq.seq) for seq in items_as_seq_records
                ]
            if item_vs_position_array is not None:
                clustering.fit(item_vs_position_array)
            else:
                raise Exception('No data available to perform clustering')
    
    assert_equal(
        ['len(items_as_seq_features)','clustering.n_leaves_'],
        locals=locals(),
        error_template_fun=(
            lambda values: f"Clustering returned wrong numer of leaves: {values[1][1]} instead of {values[0][1]}"
        )
    )
    
    aggregation_result = SimplePhylogenyWithDistances(
        num_leaves=clustering.n_leaves_,
        children=[[int(subclade) for subclade in subclades] for subclades in clustering.children_],
        max_distances=[int(distance) for distance in clustering.distances_] if hasattr(clustering, 'distances_') else None
    )

    compacted_result = compact_phylogeny(aggregation_result)
    simple_phylogeny = distances_to_branch_lengths(compacted_result)
    
    if sort:
        sort_by_leaf_indexes(simple_phylogeny)
        
    phylogeny = build_phylogeny(
        simple_phylogeny,
        items_as_seq_features = items_as_seq_features
    )
    
    return ClusteringToPhylogenyResult(
        phylogeny=phylogeny,
        item_position_to_leaf_index=simple_phylogeny.get_leaves_in_order()
    )




import csv
from Bio.SeqFeature import SeqFeature, SimpleLocation
from Bio.SeqRecord import SeqRecord
import numpy as np
from Bio.Phylo.PhyloXML import Sequence


def BED_file_to_features(BED_filename):
    with open(BED_filename) as tsv:
        return [
            SeqFeature(
                SimpleLocation(
                    int(bedLine['chromStart']),
                    int(bedLine['chromEnd']),
                    strand=(1 if bedLine['strand'] == '+' else (-1 if bedLine['strand']
                            == '-' else None)) if 'strand' in bedLine else None,
                    ref=bedLine['chrom']
                ),
                id=bedLine['name'], type='repeat'
            )
            for bedLine in csv.DictReader(tsv, delimiter="\t", fieldnames=[
                'chrom', 'chromStart', 'chromEnd', 'name', 'score', 'strand',
                'thickStart', 'thickEnd', 'itemRgb', 'blockCount', 'blockSizes', 'blockStarts'])
        ]


def feature_to_seq(
    feature: SeqFeature,
    references: dict[SeqRecord]
) -> SeqRecord:
    seq = feature.extract(None, references=references)
    seq.id = feature.id
    seq.description = ''
    return seq


def id_from_location(location: SimpleLocation) -> str:
    return (
        f'{location.ref}:{location.start}-{location.end}'
        + f'({"-" if location.strand == -1 else "+"})' if location.strand is not None else ''
    )


def location_to_feature(location: SimpleLocation) -> SeqFeature:
    return SeqFeature(location=location, id=id_from_location(location))


def location_to_seq(
    location: SimpleLocation,
    references: dict[SeqRecord]
) -> SeqRecord:
    return feature_to_seq(location_to_feature(location), references)


def label_to_location(seq_id: str) -> SimpleLocation:
    coordinates_parts = seq_id.split(':')
    seq_label = coordinates_parts[0]
    start_end = coordinates_parts[1].split('-')
    start = int(start_end[0])
    end = int(start_end[1])
    return SimpleLocation(start, end, strand=1, ref=seq_label)


def label_to_feature(seq_id: str) -> SeqFeature:
    return SeqFeature(location=label_to_location(seq_id), id=seq_id)


def label_to_phyloxml_sequence(seq_id: str) -> Sequence:
    return Sequence(name=seq_id, location=label_to_location(seq_id))


def extract_indices(indexed_list):
    return [indexed_item[0] for indexed_item in indexed_list]


def sorted_locations_indices(
    locations,
    independent_strands: bool = True
):
    indexed_locations = list(enumerate([location for location in locations]))
    sorted_indices = []
    for curr_ref in sorted(set([location.ref for location in locations])):
        curr_ref_indexed_locations = [
            indexed_location for indexed_location in indexed_locations if indexed_location[1].ref == curr_ref
        ]

        def indexed_locations_by_strand(strand):
            return [
                indexed_location
                for indexed_location in curr_ref_indexed_locations
                if indexed_location[1].strand == strand
            ]

        def get_start(indexed_location):
            return indexed_location[1].start
        
        if independent_strands:
            sorted_indices.extend(
                extract_indices(
                    sorted(
                        indexed_locations_by_strand(1),
                        key=get_start
                    )
                ) +
                extract_indices(
                    sorted(
                        indexed_locations_by_strand(-1),
                        key=get_start,
                        reverse=True
                    )
                )
            )
        else:
            sorted_indices.extend(
                extract_indices(
                    sorted(
                        curr_ref_indexed_locations,
                        key=get_start
                    )
                )

            )
    return sorted_indices


def order_by_indices(list_to_order, reordered_indeces):
    return [list_to_order[old_index] for old_index in reordered_indeces] if list_to_order is not None else None


def sorted_locations(locations):
    order_by_indices(locations, sorted_locations_indices(locations))


def sorted_features(features):
    order_by_indices(features, sorted_locations_indices(
        [feature.location for feature in features]))


def indeces_to_ordering_matrix(reordered_indeces):
    list_size = len(reordered_indeces)
    return np.array([[i == old_index for i in range(list_size)] for old_index in reordered_indeces]).T


def order_matrix_by_indeces(matrix_to_order, reordered_indeces, reorder_rows=True, reorder_cols=True):
    if matrix_to_order is None:
        return None
    return np.array([
        [
            matrix_to_order[
                reordered_indeces[row_index] if reorder_rows else row_index
            ][
                reordered_indeces[col_index] if reorder_cols else col_index
            ]
            for col_index in range(matrix_to_order.shape[1])
        ] for row_index in range(matrix_to_order.shape[0])]
    )

class SeqFeaturesByContiguity:
    sorted_seq_features: list[SeqFeature]
    gap_indices: list[int]
    max_allowed_gap : int
    reordered_indices: list[int]
    independent_strands: bool
    
    def __init__(
        self,
        seq_features: list[SeqFeature] = None,
        seq_locations: list[SimpleLocation] = None,
        seq_labels: list[str] = None,
        max_allowed_gap: int = 10,
        sorted_by_positive_strand_location: bool = False,
        sorted: bool = False,
        independent_strands: bool = True
    ) -> None:
        
        self.independent_strands = independent_strands
        
        if (
            seq_labels is not None and
            seq_features is None and seq_locations is None
        ):
            seq_features = [label_to_feature(label) for label in seq_labels]
        
        if seq_features is not None and seq_locations is None:
            seq_locations = [
                feature.location for feature in seq_features]
            
        self.max_allowed_gap = max_allowed_gap

        if seq_locations is None:
            raise Exception('Sequence location information is not available')

        if seq_features is None:
            seq_features = [
                location_to_feature(location)
                for location in seq_locations
        ]

        if not sorted and (independent_strands or not sorted_by_positive_strand_location):
            if sorted_by_positive_strand_location:
                indexed_locations = list(enumerate(self.seq_locations))
                positive_location_indices = [
                    indexed_location[0] for indexed_location in indexed_locations
                    if indexed_location[1].strand is None or indexed_location[1].strand == 1
                ]
                negative_location_indices = list(reversed([
                    indexed_location[0] for indexed_location in indexed_locations
                    if indexed_location[1].strand is not None and indexed_location[1].strand == -1
                ]))
                self.reordered_indices = positive_location_indices + negative_location_indices
            else:
                self.reordered_indices = sorted_locations_indices(
                    seq_locations,
                    independent_strands=independent_strands
                )
            self.sorted_seq_features = order_by_indices(
                seq_features,
                self.reordered_indices
            )
            
        sorted_seq_locations = [
            seq_feature.location
            for seq_feature in self.sorted_seq_features
        ]

        def locations_gap(location_1, location_2):
            return max(
                location_1.start - location_2.end,
                location_2.start - location_1.end,
            )

        self.gap_indices = [
            i + 1
            for i in range(len(sorted_seq_locations) - 1)
            if sorted_seq_locations[i + 1].ref != sorted_seq_locations[i].ref
                or (
                    independent_strands and
                    sorted_seq_locations[i + 1].strand != sorted_seq_locations[i].strand
                )
                or locations_gap(sorted_seq_locations[i], sorted_seq_locations[i + 1]) > self.max_allowed_gap
        ]
        
def feature_len(feature: SeqFeature) -> int:
    return feature.location.end - feature.location.start   

def remove_overlapping_features(
    features: list[SeqFeature],
    max_allowed_overlap: int = None,
    max_allowed_overlap_fraction: float = None,
    expected_feature_size: int = None,
    keep_overlap_warning: bool = True,
    remove_overlap_warning: bool = True
) -> list[SeqFeature]:
    if (
        max_allowed_overlap is None and
        max_allowed_overlap_fraction is not None and
        expected_feature_size is not None
    ):
        max_allowed_overlap = expected_feature_size * max_allowed_overlap_fraction
    if max_allowed_overlap is None and not keep_overlap_warning:
        return
    filtered_features: list[SeqFeature] = []
    sorted_features = sorted(features, key=lambda ref: (ref.ref, ref.location.start))
    for curr_feature in sorted_features:
        prev_feature = filtered_features[-1] if len(filtered_features) > 0 else None
        if prev_feature is not None and curr_feature.ref == prev_feature.ref and curr_feature.location.start < prev_feature.location.end:
            if prev_feature.location.end - curr_feature.location.start > max_allowed_overlap:
                keep_previous = (
                    (
                        expected_feature_size is None and
                        feature_len(prev_feature) > feature_len(curr_feature)
                    ) or (
                        abs(feature_len(prev_feature) - expected_feature_size) <=
                        abs(feature_len(curr_feature) - expected_feature_size)
                    )
                )
                if not keep_previous:
                    filtered_features[-1] = curr_feature
                if remove_overlap_warning:
                    print(
                        f"Removed sequence feature {(curr_feature if keep_previous else prev_feature).location} " +
                        f"overlapping ({prev_feature.location.end - curr_feature.location.start} bps) " +
                        f"with sequence feature {(prev_feature if keep_previous else curr_feature).location}."
                    )
            else:
                filtered_features.append(curr_feature)
                if keep_overlap_warning:
                    print(
                        f"Kept sequence feature {curr_feature.location} " +
                        f"overlapping ({prev_feature.location.end - curr_feature.location.start} bps) " +
                        f"with sequence feature {prev_feature.location}."
                    )
        else:
            filtered_features.append(curr_feature)
    return filtered_features

def get_seq_as_txt(seq: list[int]) -> str:
    return "".join([(chr(65 + symbol) if symbol < 26 else (chr(71 + symbol) if symbol < 52 else '*')) for symbol in seq])



from dataclasses import dataclass
import re
from typing import Optional
from Bio.Phylo.BaseTree import Tree, Clade
from Bio.SeqFeature import SimpleLocation
from .featureUtils import SeqFeaturesByContiguity, label_to_feature, location_to_feature

@dataclass
class CladesByLevelResult:
    clades_by_level: list[list[Clade]]
    children_num_by_level: list[list[int]]
    
class CladeHeights:
    __clade_heights_dict: dict[Clade, int]
    tree_height: int
    discrete_sorting: bool
    
    def calc_clade_discrete_heights(self, phylogeny: Tree):
        
        self.__clade_heights_dict = {}

        def calc_clade_height(clade: Clade) -> int:
            clade_height = (
                0 if len(clade.clades) == 0
                else 1 + max([
                    calc_clade_height(subclade)
                    for subclade in clade.clades
                ])
            )
            self.__clade_heights_dict[clade] = clade_height
            return clade_height
        
        self.tree_height = calc_clade_height(phylogeny.root)
    
    def calc_clade_weighted_heights(self, phylogeny: Tree):
        clade_weighted_heights_list: list[tuple[float, Clade]] = []
        
        def calc_clade_weighted_height(clade: Clade) -> int:
            # leafs, -1.0 is to avoid clash with potential level 0.0 of nodes joining identical sequences
            if len(clade.clades) == 0:
                clade_weighted_heights_list.append((-1.0, clade))
                return 0.0
                
            clade_height = max([
                subclade.branch_length + calc_clade_weighted_height(subclade)
                for subclade in clade.clades
            ])
            clade_weighted_heights_list.append((clade_height, clade))
            return clade_height
        
        calc_clade_weighted_height(phylogeny.root)

        self.__clade_heights_dict = {}
        discrete_height = 0
        curr_weighted_height = -1.0
        for (weighted_height, clade) in sorted(
            clade_weighted_heights_list,
            key=lambda height_and_clade:height_and_clade[0]
        ):
            if weighted_height > curr_weighted_height:
                discrete_height += 1
            self.__clade_heights_dict[clade] = discrete_height
        self.tree_height = discrete_height
    
    def __init__(self, phylogeny: Tree, discrete_sorting: bool = False) -> None:
        self.discrete_sorting = discrete_sorting
        if discrete_sorting:
            self.calc_clade_discrete_heights(phylogeny)
        else:
            self.calc_clade_weighted_heights(phylogeny)

    def get_clade_height(self, clade: Clade) -> int:
        return self.__clade_heights_dict[clade]

def get_clades_by_level(
    phylogeny: Tree,
    discrete_sorting: bool = False
) -> CladesByLevelResult:
    
    clades_by_level: list[list[Clade]] = []
    children_num_by_level: list[list[int]] = []

    # def get_clade_height(clade: Clade) -> int:
    #     return (
    #         0 if len(clade.clades) == 0
    #         else 1 + max([
    #             get_clade_height(subclade)
    #             for subclade in clade.clades
    #         ])
    #     )
    
    clade_heights = CladeHeights(
        phylogeny=phylogeny,
        discrete_sorting=discrete_sorting
    )

    def set_by_height(clade_height: int, clade: Clade, children_num: int):
        if clade_height >= len(clades_by_level):
            missing_levels = clade_height - len(clades_by_level) + 1
            clades_by_level.extend([[]] * missing_levels)
            children_num_by_level.extend([[]] * missing_levels)
        clades_by_level[clade_height].append(clade)
        children_num_by_level[clade_height].append(children_num)

    def extract_clades_by_height(clade: Clade):
        subclades = clade.clades
        subclade_heights = [
            clade_heights.get_clade_height(subclade)
            for subclade in subclades
        ]
        # clade_height = (
        #     0 if len(subclades) == 0
        #     else 1 + max(subclade_heights)
        # )
        clade_height = clade_heights.get_clade_height(clade)
        for subclade_position, subclade_height in enumerate(subclade_heights):
            subclade = subclades[subclade_position]
            extract_clades_by_height(subclade)
            for height in range(subclade_height + 1, clade_height):
                set_by_height(clade_height=height, clade=subclades[subclade_position], children_num=1)            
        set_by_height(clade_height=clade_height, clade=clade, children_num=len(subclades))

    extract_clades_by_height(phylogeny.root)
    return CladesByLevelResult(
        clades_by_level=clades_by_level,
        children_num_by_level=children_num_by_level[1:]
    )

def get_clade_contraction_by_level(
    children_num_by_level: list[list[int]]
) -> list[list[int]]:
    contraction_by_level = []
    for children_num_by_clade in children_num_by_level:
        contraction = []
        for clade_index, children_num in enumerate(children_num_by_clade):
            contraction.extend([clade_index] * children_num)
        contraction_by_level.append(contraction)
    return contraction_by_level
        
def get_labelled_items_by_level(
    clade_contraction_by_level: list[list[int]],
    item_position_to_leaf_index: list[int]
) -> list[list[int]]:
    labelled_items_by_level = [item_position_to_leaf_index]
    for level, contraction in enumerate(clade_contraction_by_level):
        labelled_items_by_level.append([
            contraction[label]
            for label in labelled_items_by_level[level]
        ])
    return labelled_items_by_level

location_pattern = re.compile("(.*)\[([0-9]+):([0-9]+)\]\((\+|-)\)")

def parse_location(location_str: str) -> SimpleLocation:
    match = location_pattern.match(location_str)
    return SimpleLocation(
        ref=match.group(1),
        start=int(match.group(2)),
        end=int(match.group(3)),
        strand=int(match.group(4) + '1'))

def extract_features_from_leaves(phylogeny: Tree):
    return [
        location_to_feature(
            parse_location(leave.sequences[0].location)
            if isinstance(leave.sequences[0].location, str)
            else leave.sequences[0].location
        )
        if (
            hasattr(leave, 'sequences') and
            len(leave.sequences) == 1 and
            hasattr(leave.sequences[0], 'location')
        )
        else label_to_feature(leave.name)
        for leave in phylogeny.get_terminals()
    ]

@dataclass
class PhylogenyToLevelsResult:
    clades_by_level: list[list[Clade]]
    clade_contraction_by_level: list[list[int]]
    labelled_items_by_level: list[list[int]]
    
def phylogeny_to_levels(
    phylogeny: Tree,
    item_position_to_leaf_index: Optional[list[int]] = None,
    max_allowed_gap: int = 10,
    discrete_sorting: bool = False
) -> PhylogenyToLevelsResult:
    if item_position_to_leaf_index is None:
        item_position_to_leaf_index = SeqFeaturesByContiguity(
            seq_features=extract_features_from_leaves(phylogeny),
            max_allowed_gap=max_allowed_gap
        ).reordered_indices
    clades_by_level_res = get_clades_by_level(
        phylogeny,
        discrete_sorting=discrete_sorting
    )
    clade_contraction_by_level = get_clade_contraction_by_level(
        clades_by_level_res.children_num_by_level
    )
    labelled_items_by_level = get_labelled_items_by_level(
        clade_contraction_by_level=clade_contraction_by_level,
        item_position_to_leaf_index=item_position_to_leaf_index
    )
    return PhylogenyToLevelsResult(
        clades_by_level=clades_by_level_res.clades_by_level,
        clade_contraction_by_level=clade_contraction_by_level,
        labelled_items_by_level=labelled_items_by_level
    )


from typing_extensions import Callable

def default_error_template_fun(values: list[tuple[str,int]]) -> str:
    values_str = '\n'.join(map(
        lambda expr_value: f"{expr_value[0]} = {expr_value[1]}",
        values
    ))
    return f"Inconsistent values found:\n{values_str}"

def assert_equal(
    exprs: list[str],
    locals: dict,
    error_template_fun: Callable[[list[tuple[str,int]]], str] = default_error_template_fun
):
    values_to_check: list[tuple[str,int]] = []
    for expr in exprs:
        try:
            values_to_check.append((expr, eval(expr, None, locals)))
        except:
            pass
        
    common_value = None
    for (expr, value) in values_to_check:
        if common_value is None:
            common_value = value
        else:
            if value != common_value:
                raise Exception(error_template_fun(values_to_check))
